/* session.c, session management routines originally designed for OpenTAS

 * Copyright (C) 2001,2002 M.C. Bezemer
 *
 * Licensed under the terms of the GNU General Public License.
 * ABSOLUTELY NO WARRANTY.
 * See the file `COPYING' in this directory.
 */

#include <stdio.h>
#include <lockfile.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "urlencode.h"
#include "urldecode.h"
#include "findvalue.h"
#include "session.h"

#ifndef LOCKFILE
#define LOCKFILE "/home/opentas/opentas.lock"
#endif
#ifndef SESSIONFILE
#define SESSIONFILE "/home/opentas/opentas.sessions"
#endif
#define LINELEN 2048
#define TIMEOUT 600

int save_session_data(char *username, char *password, 
		      char *tas_stud_id, char **ses_id)
/* saves username, password and tas_stud_id in the session file and
 * returns the session id in ses_id
 */
{
	FILE *fp=NULL;
	FILE *urandom=NULL;
	int i=0;
	int j=0;
	char *session_id=NULL;
	unsigned char *ses_bin=NULL;

	if ((username==NULL) || (password==NULL) || (tas_stud_id==NULL) ) {
		return -1;
	}

	urandom=fopen("/dev/urandom","r");
	if (urandom==NULL) {
		return 1;
	}

	ses_bin=malloc(17*sizeof(char));
	if (ses_bin==NULL) {
		return 2;
	}
	session_id=malloc(33*sizeof(char));
	if (session_id==NULL) {
		return 3;
	}

	i=0;
	while (i<5) {
		if (fread(ses_bin,1,16,urandom)!=16) {
			free(ses_bin);
			ses_bin=NULL;
			free(session_id);
			session_id=NULL;
			return 4;
		}
		ses_bin[16]='\0';
		for (j=0;j<16;j++) {
			snprintf(session_id+2*j,3,"%02X",ses_bin[j]);
		}
		if (session_exists(session_id)==0) {
			i=0;
			break;
		}
		i++;
	}
	free(ses_bin);
	ses_bin=NULL;
	fclose(urandom);
	if (i!=0) {
		free(session_id);
		session_id=NULL;
		return 5;
	}

	i=0;
	while (i<8) {
		j=lockfile_create(LOCKFILE,1,L_PID);
		if (j==L_SUCCESS) {
			i=0;
			break;
		}
		i++;
		usleep(250000);
	}
	if (j!=L_SUCCESS) {
		free(session_id);
		session_id=NULL;
		return 6;
	}

	fp=fopen(SESSIONFILE,"a");
	if (fp==NULL) {
		free(session_id);
		session_id=NULL;
		return 7;
	}
	fprintf(fp,"session_id=%s",session_id);
	fprintf(fp,"&username=%s",urlencode(username));
	fprintf(fp,"&password=%s",urlencode(password));
	fprintf(fp,"&tas_stud_id=%s",urlencode(tas_stud_id));
	fprintf(fp,"&timestamp=%lu\n",time(NULL));
	fclose(fp);
	lockfile_remove(LOCKFILE);
	if (ses_id!=NULL) {
		*ses_id=session_id;
	} else {
		free(session_id);
		session_id=NULL;
	}

	return 0;
}


int get_session_data(char *ses_id, char **username, char **password,
		     char **tas_stud_id)
/* reads sessions file and returns username, password and tas_stud_id for the
 * given session id, or returns !0 in case of error, not found or timed-out
 */
{
	int i=0;
	int j=0;
	FILE *fp=NULL;
	FILE *fp_new=NULL;
	char line[LINELEN];
	char *read_ses_id=NULL;
	char *user=NULL;
	char *pass=NULL;
	char *studid=NULL;
	char *timestamp=NULL;
	int found=0;

	if (ses_id==NULL) {
		return -1;
	}
	i=0;
	while (i<8) {
		j=lockfile_create(LOCKFILE,1,L_PID);
		if (j==L_SUCCESS) {
			i=0;
			break;
		}
		i++;
		usleep(250000);
	}
	if (j!=L_SUCCESS) {
		return 1;
	}

	fp=fopen(SESSIONFILE,"r");
	if (fp==NULL) {
		lockfile_remove(LOCKFILE);
		return 2;
	}
	fp_new=fopen(SESSIONFILE".new","w");
	if (fp_new==NULL) {
		fclose(fp);
		lockfile_remove(LOCKFILE);
		return 3;
	}
	while ( (!feof(fp)) && (fgets(line,sizeof(line),fp)!=NULL) ) {
		if (strlen(line)==0) {
			continue;
		}
		if (line[strlen(line)-1]!='\n') {
			while ((!feof(fp)) && (fgetc(fp)!='\n'));
			continue;
		}
		inittasresponse(line);
		if ((read_ses_id=findfirstvalue("session_id"))!=NULL) {
			if (strcmp(read_ses_id,ses_id)==0) {
				if (found!=0) {
					continue;
				}
				user=findfirstvalue("username");
				pass=findfirstvalue("password");
				studid=findfirstvalue("tas_stud_id");
				timestamp=findfirstvalue("timestamp");
				if ( (user==NULL) ||
				     (pass==NULL) ||
				     (studid==NULL) ||
				     (timestamp==NULL) ) {
					continue;
				}
				if (strtoul(timestamp,NULL,10)+TIMEOUT<time(NULL)) {
					found=2;
					continue;
				}
				user=strdup(user);
				pass=strdup(pass);
				studid=strdup(studid);
				fprintf(fp_new,"session_id=%s",read_ses_id);
				fprintf(fp_new,"&username=%s",urlencode(user));
				fprintf(fp_new,"&password=%s",urlencode(pass));
				fprintf(fp_new,"&tas_stud_id=%s",urlencode(studid));
				fprintf(fp_new,"&timestamp=%lu\n",time(NULL));
				found=1;
			} else {
				fprintf(fp_new,"%s",line);
			}
		}
	}
	fclose(fp);
	fclose(fp_new);
	rename(SESSIONFILE".new",SESSIONFILE);

	lockfile_remove(LOCKFILE);

	if (found==2) {
		return -2;
	}
	if (found!=1) {
		return -3;
	}
	if (username!=NULL) {
		*username=user;
	}
	if (password!=NULL) {
		*password=pass;
	}
	if (tas_stud_id!=NULL) {
		*tas_stud_id=studid;
	}

	return 0;
}


int change_session_password(char *ses_id, char *newpassword)
/* changes password of the given session, or returns !0 in case of error,
 * not found or timed-out
 */
{
	int i=0;
	int j=0;
	FILE *fp=NULL;
	FILE *fp_new=NULL;
	char line[LINELEN];
	char *read_ses_id=NULL;
	char *user=NULL;
	char *pass=NULL;
	char *studid=NULL;
	char *timestamp=NULL;
	int found=0;

	if (ses_id==NULL) {
		return -1;
	}
	i=0;
	while (i<8) {
		j=lockfile_create(LOCKFILE,1,L_PID);
		if (j==L_SUCCESS) {
			i=0;
			break;
		}
		i++;
		usleep(250000);
	}
	if (j!=L_SUCCESS) {
		return 1;
	}

	fp=fopen(SESSIONFILE,"r");
	if (fp==NULL) {
		lockfile_remove(LOCKFILE);
		return 2;
	}
	fp_new=fopen(SESSIONFILE".new","w");
	if (fp_new==NULL) {
		fclose(fp);
		lockfile_remove(LOCKFILE);
		return 3;
	}
	while ( (!feof(fp)) && (fgets(line,sizeof(line),fp)!=NULL) ) {
		if (strlen(line)==0) {
			continue;
		}
		if (line[strlen(line)-1]!='\n') {
			while ((!feof(fp)) && (fgetc(fp)!='\n'));
			continue;
		}
		inittasresponse(line);
		if ((read_ses_id=findfirstvalue("session_id"))!=NULL) {
			if (strcmp(read_ses_id,ses_id)==0) {
				if (found!=0) {
					continue;
				}
				user=findfirstvalue("username");
				pass=findfirstvalue("password");
				studid=findfirstvalue("tas_stud_id");
				timestamp=findfirstvalue("timestamp");
				if ( (user==NULL) ||
				     (pass==NULL) ||
				     (studid==NULL) ||
				     (timestamp==NULL) ) {
					continue;
				}
				if (strtoul(timestamp,NULL,10)+TIMEOUT<time(NULL)) {
					found=2;
					continue;
				}
				fprintf(fp_new,"session_id=%s",read_ses_id);
				fprintf(fp_new,"&username=%s",urlencode(user));
				fprintf(fp_new,"&password=%s",urlencode(newpassword));
				fprintf(fp_new,"&tas_stud_id=%s",urlencode(studid));
				fprintf(fp_new,"&timestamp=%lu\n",time(NULL));
				found=1;
			} else {
				fprintf(fp_new,"%s",line);
			}
		}
	}
	fclose(fp);
	fclose(fp_new);
	rename(SESSIONFILE".new",SESSIONFILE);

	lockfile_remove(LOCKFILE);

	if (found==2) {
		return -2;
	}
	if (found!=1) {
		return -3;
	}
	return 0;
}


int destroy_session(char *ses_id)
/* removes the session given in ses_id from the sessions file
 * returns 0 if ok, !0 else.
 */
{
	int i=0;
	int j=0;
	FILE *fp=NULL;
	FILE *fp_new=NULL;
	char line[1000];
	char *read_ses_id=NULL;

	if (ses_id==NULL) {
		return -1;
	}
	i=0;
	while (i<8) {
		j=lockfile_create(LOCKFILE,1,L_PID);
		if (j==L_SUCCESS) {
			i=0;
			break;
		}
		i++;
		usleep(250000);
	}
	if (j!=L_SUCCESS) {
		return 1;
	}

	fp=fopen(SESSIONFILE,"r");
	if (fp==NULL) {
		return 2;
	}
	fp_new=fopen(SESSIONFILE".new","w");
	if (fp_new==NULL) {
		return 3;
	}
	while ( (!feof(fp)) && (fgets(line,sizeof(line),fp)!=NULL) ) {
		if (strlen(line)==0) {
			continue;
		}
		if (line[strlen(line)-1]!='\n') {
			while ((!feof(fp)) && (fgetc(fp)!='\n'));
			continue;
		}
		inittasresponse(line);
		if ((read_ses_id=findfirstvalue("session_id"))!=NULL) {
			if (strcmp(read_ses_id,ses_id)==0) {
				continue;
			} else {
				fprintf(fp_new,"%s",line);
			}
		}
	}
	fclose(fp);
	fclose(fp_new);
	rename(SESSIONFILE".new",SESSIONFILE);

	lockfile_remove(LOCKFILE);

	return 0;
}

int session_exists(char *ses_id)
//returns 0 if session doesn't exist, !0 if it does or in case of error
{
	int i=0;
	int j=0;
	FILE *fp=NULL;
	char line[1000];
	char *read_ses_id=NULL;
	int found=0;

	if (ses_id==NULL) {
		return -1;
	}
	i=0;
	while (i<8) {
		j=lockfile_create(LOCKFILE,1,L_PID);
		if (j==L_SUCCESS) {
			i=0;
			break;
		}
		i++;
		usleep(250000);
	}
	if (j!=L_SUCCESS) {
		return 2;
	}

	fp=fopen(SESSIONFILE,"r");
	if (fp==NULL) {
		lockfile_remove(LOCKFILE);
 		/* If sessionsfile doesn't exist, it's not likely to
		 * contain any session... so return 0 instead of error
		 */
		lockfile_remove(LOCKFILE);
		return 0;
	}
	while ( (!feof(fp)) && (fgets(line,sizeof(line),fp)!=NULL) ) {
		if (strlen(line)==0) {
			continue;
		}
		if (line[strlen(line)-1]!='\n') {
			while ((!feof(fp)) && (fgetc(fp)!='\n'));
			continue;
		}
		inittasresponse(line);
		if ((read_ses_id=findfirstvalue("session_id"))!=NULL) {
			if (strcmp(read_ses_id,ses_id)==0) {
				found=1;
				break;
			}
		}
	}
	fclose(fp);

	lockfile_remove(LOCKFILE);

	if (found!=0) {
		return 1;
	}

	return 0;
}

int cleanup_sessionfile()
// drops illegal lines and timed-out sessions from session file
//
// returns:
//  amount of dropped sessions (zero or positive) or
//  error code (negative)
{
	int i=0;
	int j=0;
	FILE *fp=NULL;
	FILE *fp_new=NULL;
	char line[1000];
	char *read_ses_id=NULL;
	char *timestamp=NULL;
	int amount_dropped=0;

	i=0;
	while (i<8) {
		j=lockfile_create(LOCKFILE,1,L_PID);
		if (j==L_SUCCESS) {
			i=0;
			break;
		}
		i++;
		usleep(250000);
	}
	if (j!=L_SUCCESS) {
		return -2;
	}

	fp=fopen(SESSIONFILE,"r");
	if (fp==NULL) {
		lockfile_remove(LOCKFILE);
		return -3;
	}
	fp_new=fopen(SESSIONFILE".new","w");
	if (fp_new==NULL) {
		fclose(fp);
		lockfile_remove(LOCKFILE);
		return -4;
	}
	amount_dropped=0;
	while ( (!feof(fp)) && (fgets(line,sizeof(line),fp)!=NULL) ) {
		if (strlen(line)==0) {
			amount_dropped++;
			continue;
		}
		if (line[strlen(line)-1]!='\n') {
			while ((!feof(fp)) && (fgetc(fp)!='\n'));
			amount_dropped++;
			continue;
		}
		inittasresponse(line);
		if ((read_ses_id=findfirstvalue("session_id"))!=NULL) {
			timestamp=findfirstvalue("timestamp");
			if (timestamp==NULL) {
				amount_dropped++;
				continue;
			}
			if (strtoul(timestamp,NULL,10)+TIMEOUT<time(NULL)) {
				amount_dropped++;
				continue;
			}
			fprintf(fp_new,"%s",line);
		} else {
			amount_dropped++;
		}
	}
	fclose(fp);
	fclose(fp_new);
	rename(SESSIONFILE".new",SESSIONFILE);

	lockfile_remove(LOCKFILE);

	return amount_dropped;

	return 0;
}
