/* tasquery.c, compose and send query to TAS server
   
 * Copyright (C) 2001,2002 M.C. Bezemer
 *
 * Largely based on tascommand2.c by Joris van Rantwijk
 *
 * Licensed under the terms of the GNU General Public License.
 * ABSOLUTELY NO WARRANTY.
 * See the file `COPYING' in this directory.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>

#include "tasquery.h"
#include "urlencode.h"

int tq_alarmed=0;
char *tq_tasquerystring=NULL;
size_t tq_tasquerystring_strlen=0;
size_t tq_tasquerystring_buflen=0;


void tas_query_alarm(int signal)
/* alarm handler to give up on unresponsive remote hosts
 */
{
	if (signal==SIGALRM) {
		tq_alarmed=1;
	}
	return;
}


int tas_query_init(void)
/* resets the tasquerystring buffer
 * returns 0 if ok, !0 if error
 */
{
	tq_tasquerystring_buflen=1000;
	tq_tasquerystring_strlen=0;
	tq_tasquerystring=realloc(tq_tasquerystring,
				 tq_tasquerystring_buflen*sizeof(char));
	if (tq_tasquerystring==NULL) {
		return -1;
	}
	tq_tasquerystring[0]='\0';
	return 0;
}


int tas_query_add(char *string, int urlencoded)
/* adds the given string to the tasquerystring buffer
 * uses urlencode if urlencoded !=0
 * returns 0 if ok, !0 on error
 */
{
	size_t stringlen=0;
	char *addstring=NULL;

	if (string==NULL) {
		return 0;
	}
	if (urlencoded) {
		addstring=urlencode(string);
	} else {
		addstring=string;
	}
	if (addstring!=NULL) {
		stringlen=strlen(addstring);
	} else {
		stringlen=0;
	}
	if (tq_tasquerystring_strlen+stringlen+1 > tq_tasquerystring_buflen) {
		tq_tasquerystring_buflen=tq_tasquerystring_strlen+stringlen+1000;
		tq_tasquerystring=realloc(tq_tasquerystring,tq_tasquerystring_buflen);
	}
	if (tq_tasquerystring==NULL) {
		return -2;
	}
	memcpy(tq_tasquerystring+tq_tasquerystring_strlen,addstring,stringlen);

	tq_tasquerystring_strlen+=stringlen;
	tq_tasquerystring[tq_tasquerystring_strlen]='\0';

	return 0;
}


int add_to_buf(char **buf, int *buflen, int *bufsize, char *addition, int addition_len)
{
	if (addition_len==0) {
		if (addition!=NULL) {
			addition_len=strlen(addition);
		}
	}
	if (*bufsize < *buflen+addition_len+1) {
		*bufsize+=addition_len+10000;
		*buf=realloc(*buf,*bufsize);
		if (*buf==NULL) {
			return -1;
		}
	}
	memcpy(*buf+*buflen,addition,addition_len);
	*buflen += addition_len;
	(*buf)[*buflen]='\0';
	return 0;
}


int do_tas_query(char **result, int *result_len)
{
	return tas_query(tq_tasquerystring,
			 tq_tasquerystring_strlen,
			 result,
			 result_len);
}


int tas_query(char *query, int query_len, char **result, int *result_len)
{
	int targetsock;
	struct sockaddr_in sin;
	struct hostent *hp;
	struct sigaction act;
	struct sigaction old_act;
	
	SSL_CTX *ssl_ctx = NULL;
	SSL *ssl;
	BIO *ssl_bio;
	BIO *out;
	char buf[16000];
	char *p;
	int i;
	int len;
	int offs;
	static char *res=NULL;
	int ressize=0;
	int reslen=0;
	int result_size=0;

	if (res!=NULL) {
		free(res);
		res=NULL;
		ressize=0;
		reslen=0;
	}

	*result_len=0;
	*result=NULL;

	sigaction(SIGALRM,NULL,&act);
	memcpy(&old_act,&act,sizeof(act));
	act.sa_handler=tas_query_alarm;
	act.sa_flags &= ~SA_RESTART;
	sigaction(SIGALRM,&act,NULL);
	tq_alarmed=0;

	/* connect to the TAS https server */
	/* first set local IP for connection */
	targetsock=socket(AF_INET, SOCK_STREAM, 0);
	if (targetsock < 2) {
		add_to_buf(result,result_len,&result_size,
			   "socket() failed",0);
		sigaction(SIGALRM,&old_act,NULL);
		return -1;
	}
	sin.sin_family=AF_INET;
	sin.sin_port=0;
	hp=gethostbyname(LOCAL_HOST);
	if (hp==NULL) {
		add_to_buf(result,result_len,&result_size,
			   "gethostbyname() failed",0);
		sigaction(SIGALRM,&old_act,NULL);
		return -1;
	}
	memcpy(&sin.sin_addr,hp->h_addr,hp->h_length);
	if (bind(targetsock, (struct sockaddr *) &sin, sizeof(sin))<0) {
		add_to_buf(result,result_len,&result_size,
			   "bind() failed",0);
		sigaction(SIGALRM,&old_act,NULL);
		return -1;
	}
	/* next, set destination IP */
	sin.sin_family = AF_INET;

	sin.sin_port = htons(TAS_PORT);
	hp=gethostbyname(TAS_HOST);
	if (hp==NULL) {
		add_to_buf(result,result_len,&result_size,
			   "gethostbyname() failed",0);
		sigaction(SIGALRM,&old_act,NULL);
		return -1;
	}
	memcpy(&sin.sin_addr,hp->h_addr,hp->h_length);

	alarm(5);
	if (connect(targetsock, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
		if (tq_alarmed!=0) {
			add_to_buf(result,result_len,&result_size,
				   "connect() timed out",0);
			sigaction(SIGALRM,&old_act,NULL);
			return -1;
		}
		add_to_buf(result,result_len,&result_size,
			   "connect() failed out",0);
		sigaction(SIGALRM,&old_act,NULL);
		return -1;
	}
	if (tq_alarmed!=0) {
		add_to_buf(result,result_len,&result_size,
			   "connect() timed out",0);
		sigaction(SIGALRM,&old_act,NULL);
		return -1;
	}
	alarm(10);

	/* Lets get nice error messages */
	SSL_load_error_strings();

	/* Setup all the global SSL stuff */
	SSL_library_init();
	ssl_ctx = SSL_CTX_new(SSLv23_client_method());

	/* Lets make a SSL structure */
	ssl = SSL_new(ssl_ctx);
	SSL_set_connect_state(ssl);

	/* Use it inside an SSL BIO */
	ssl_bio = BIO_new(BIO_f_ssl());
	BIO_set_ssl(ssl_bio, ssl, BIO_CLOSE);

	snprintf(buf, sizeof(buf),
		"POST /webtas/cgi-bin/webtas_prod.exe HTTP/1.0\r\n"
		"Referer: https://www.tas.tudelft.nl/webtas/webtas_zonder.html\r\n"
		"User-Agent: OpenTAS\r\n"
		"Pragma: no-cache\r\n"
		"Host: www.tas.tudelft.nl\r\n"
		"Content-Type: application/x-www-form-urlencoded\r\n"
		"Content-Length: %d\r\n\r\n", query_len);

	/* Lets use an fd BIO under the SSL BIO */
	out = BIO_new(BIO_s_fd());
	BIO_set_fd(out,targetsock,BIO_CLOSE);

	BIO_set_nbio(out, 0);	/* Don't use non-blocking I/O */
	out = BIO_push(ssl_bio, out);

	len = strlen(buf);
	offs = 0;
	while (offs < len) {
		i = BIO_write(out, buf + offs, len - offs);
		if (i <= 0) {
			add_to_buf(result,result_len,&result_size,
				   "HTTPS write error. ",0);
			if (ERR_peek_error() == 0) {
				add_to_buf(result,result_len,&result_size,
				strerror(errno),0);
			}
// else			ERR_print_errors_fp(stdout);
			sigaction(SIGALRM,&old_act,NULL);
			return -1;
		}
		offs += i;
		if (tq_alarmed!=0) {
			add_to_buf(result,result_len,&result_size,
				   "BIO_write() timed out",0);
			sigaction(SIGALRM,&old_act,NULL);
			return -1;
		}
	}
	if (tq_alarmed!=0) {
		add_to_buf(result,result_len,&result_size,
			   "BIO_write() timed out",0);
		sigaction(SIGALRM,&old_act,NULL);
		return -1;
	}

	offs = 0;
	while (offs < query_len) {
		i = BIO_write(out, query + offs, query_len - offs);
		if (i <= 0) {
			add_to_buf(result,result_len,&result_size,
				   "HTTPS write error. ",0);
			if (ERR_peek_error() == 0) {
				add_to_buf(result,result_len,&result_size,
				strerror(errno),0);
			}
// else			ERR_print_errors_fp(stdout);
			sigaction(SIGALRM,&old_act,NULL);
			return -1;
		}
		offs += i;
		if (tq_alarmed!=0) {
			add_to_buf(result,result_len,&result_size,
				   "BIO_write() timed out",0);
			sigaction(SIGALRM,&old_act,NULL);
			return -1;
		}
	}
	if (tq_alarmed!=0) {
		add_to_buf(result,result_len,&result_size,
			   "BIO_write() timed out",0);
		sigaction(SIGALRM,&old_act,NULL);
		return -1;
	}

	alarm(20);

	len = 0;
	while (1) {
		i = BIO_read(out, buf, sizeof(buf)-1);
		if (tq_alarmed!=0) {
			add_to_buf(result,result_len,&result_size,
				   "BIO_read() timed out",0);
			sigaction(SIGALRM,&old_act,NULL);
			return -1;
		}
		if (i == 0)
			break;
		if (i < 0) {
			add_to_buf(result,result_len,&result_size,
				   "HTTPS read error. ",0);
			if (ERR_peek_error() == 0) {
				add_to_buf(result,result_len,&result_size,
				strerror(errno),0);
			}
// else			ERR_print_errors_fp(stdout);
			sigaction(SIGALRM,&old_act,NULL);
			return -1;
		}
		add_to_buf(&res,&reslen,&ressize,buf,i);
	}

	alarm(5);

	BIO_free_all(out);
	if (ssl_ctx != NULL) SSL_CTX_free(ssl_ctx);

	p = strchr(res, '\n');
	if (p != NULL) *p++ = '\0';
	if (strstr(res, "200 OK") == NULL) {
		add_to_buf(result,result_len,&result_size,
			   "No '200 OK' in response",0);
		sigaction(SIGALRM,&old_act,NULL);
		return -1;
	}

	while (p != NULL && *p != '\0' && strncmp(p, "\r\n", 2) != 0) {
		p = strchr(p, '\n');
		if (p != NULL) p++;
	}

	if (p == NULL || *p == '\0') {
		add_to_buf(result,result_len,&result_size,
			   "Invalid message header format",0);
		alarm(0);
		sigaction(SIGALRM,&old_act,NULL);
		return -1;
	}

	p = strchr(p, '\n');
	if (p != NULL) p++;

	add_to_buf(result,result_len,&result_size,
		   p,reslen-(p-res));
	free(res);
	res=NULL;
	reslen=0;
	ressize=0;

	alarm(0);
	sigaction(SIGALRM,&old_act,NULL);
	return 0;
}
